(define-module (ordo play)
  #:use-module (oop goops)
  #:use-module (ice-9 match)
  #:use-module (logging logger)
  #:use-module (srfi srfi-1) ; list utils
  #:use-module (srfi srfi-9) ; records
  #:use-module (srfi srfi-26) ; cut
  #:use-module (ordo connection)
  #:use-module (ordo context)
  #:use-module (ordo task)
  #:use-module (ordo handler)
  #:use-module (ordo logger)
  #:export (play run-play))

(define-record-type <play>
  (make-play description connection vars tasks handlers)
  play?
  (connection play-connection)
  (vars play-vars)
  (description play-description)
  (tasks play-tasks)
  (handlers play-handlers))

(define (validate-connection connection)
  (unless (and connection (is-a? connection <connection>))
    (error (format #f "invalid connection: ~a" connection))))

(define (validate-tasks tasks)
  (unless (and tasks (not (null? tasks)) (every task? tasks))
    (error (format #f "invalid tasks: ~a" tasks))))

(define (validate-handlers handlers)
  (unless (every (lambda (h) (and (pair? h) (handler? (cdr h)))) handlers)
    (error (format #f "invalid handlers: ~a" handlers))))

(define (validate-vars vars)
  (unless (every pair? vars)
    (error (format #f "invalid vars: ~a" vars))))

(define (validate-triggers tasks handlers)
  (for-each (lambda (task)
              (for-each (lambda (trigger)
                          (unless (assoc-ref handlers trigger)
                            (error (format #f "task \"~a\" references an undefined trigger: ~a"
                                           (task-description task)
                                           trigger))))
                        (task-triggers task)))
            tasks))

(define* (play description #:key connection tasks (vars '()) (handlers '()))
  (validate-connection connection)
  (validate-tasks tasks)
  (validate-handlers handlers)
  (validate-triggers tasks handlers)
  (validate-vars vars)
  ;; Reconstruct the vars here because, when a quoted list is passed in the
  ;; play, it can result in an error (expected mutable pair) from assoc-set!
  ;; from register-context-var!.
  (make-play description connection (fold (match-lambda* (((k . v) accum) (alist-cons k v accum))) '() vars) tasks handlers))

(define (run-play play)
  ;; TODO move logging setup and shutdown to a higher level when we implement playbook etc.
  (setup-logging)
  (log-msg 'NOTICE "Running play: " (play-description play))
  (call-with-connection
   (play-connection play)
   (lambda (c)
     (let* ((ctx (make-context c (play-vars play))))
       (for-each (cut run-task ctx <>) (play-tasks play))
       (for-each (match-lambda
                   ((name . handler)
                    (when (context-triggered? ctx name)
                      (run-handler ctx handler))))
                 (play-handlers play)))))
  (log-msg 'NOTICE "Completed play: " (play-description play))
  (shutdown-logging))
